import concurrent.futures
import uuid
import os
import pathlib
from contextlib import closing
from ..classes.Row import Row
from tqdm.auto import tqdm
import shutil
from .util import get_tempdir
from .api_request import make_request
from threading import Event

MAX_PARALLELIZATION = 8


class RedivisArrowIterator:
    def __init__(self, streams, mapped_variables, progressbar, coerce_schema):
        self.streams = streams
        self.mapped_variables = mapped_variables
        self.progressbar = progressbar
        self.coerce_schema = coerce_schema
        self.current_stream_index = 0
        self.__get_next_reader__()

    def __get_next_reader__(self):
        import pyarrow

        # TODO: this won't get closed properly if the iterator is not fully consumed
        arrow_response = make_request(
            method="get",
            path=f'/readStreams/{self.streams[self.current_stream_index]["id"]}',
            stream=True,
            parse_response=False,
        )
        self.current_record_batch_reader = pyarrow.ipc.RecordBatchStreamReader(
            arrow_response.raw
        )
        if self.coerce_schema:
            self.variables_in_stream = list(
                map(
                    lambda field_name: next(
                        x for x in self.mapped_variables if x["name"] == field_name
                    ),
                    self.current_record_batch_reader.schema.names,
                )
            )
            self.output_schema = pyarrow.schema(
                map(variable_to_field, self.variables_in_stream)
            )
        else:
            self.output_schema = self.current_record_batch_reader.schema

    def __iter__(self):
        return self

    def __next__(self):
        import pyarrow

        try:
            batch = self.current_record_batch_reader.read_next_batch()
            if self.coerce_schema:
                batch = pyarrow.RecordBatch.from_arrays(
                    list(
                        map(
                            coerce_string_variable,
                            batch.columns,
                            self.variables_in_stream,
                        )
                    ),
                    schema=self.output_schema,
                )

            if self.progressbar is not None:
                self.progressbar.update(batch.num_rows)

            return batch
        except StopIteration:
            if self.current_stream_index == len(self.streams) - 1:
                if self.progressbar:
                    self.progressbar.close()
                raise StopIteration
            else:
                self.current_stream_index += 1
                self.__get_next_reader__()
                return self.__next__()


def list_rows(
    *,
    uri,
    output_type="dataframe",
    max_results=None,
    selected_variables=None,
    mapped_variables=None,
    progress=True,
    coerce_schema=False,
    batch_preprocessor=None,
    table=None,
    use_export_api=False,
    max_parallelization=os.cpu_count(),
):
    import pyarrow
    import pyarrow.dataset as pyarrow_dataset  # need to import separately, it's not on the pyarrow import

    use_export_api = (
        use_export_api
        and table
        and output_type != "arrow_iterator"
        and selected_variables is None
        and batch_preprocessor is None
        and max_results is None
    )
    progressbar = None

    if max_parallelization < 1:
        raise ValueError("max_parallelization must be greater than 0")

    pyarrow.set_cpu_count(max_parallelization)
    pyarrow.set_io_thread_count(max_parallelization)

    payload = {"requestedStreamCount": min(MAX_PARALLELIZATION, max_parallelization)}

    if max_results is not None:
        payload["maxResults"] = max_results

    if selected_variables is not None:
        payload["selectedVariables"] = selected_variables

    if not use_export_api:
        read_session = make_request(
            method="post",
            path=f"{uri}/readSessions",
            parse_response=True,
            payload=payload,
        )

        if progress:
            progressbar = tqdm(total=read_session["numRows"], leave=False)

        if output_type == "arrow_iterator":
            return RedivisArrowIterator(
                streams=read_session["streams"],
                mapped_variables=mapped_variables,
                progressbar=progressbar,
                coerce_schema=coerce_schema,
            )

    folder = pathlib.Path().joinpath(
        get_tempdir(),
        "tables",
        f"{uuid.uuid4()}",
    )

    # get the absolute folder path, as a string
    folder_path = str(folder.absolute())

    try:
        if use_export_api:
            table.download(folder_path + "/", format="parquet", progress=progress)
            arrow_dataset = pyarrow_dataset.dataset(folder_path, format="parquet")
        else:
            # create the folder, if it doesn't exist
            folder.mkdir(parents=True, exist_ok=True)
            # Use download_state to notify worker threads when to quit.
            # See: https://stackoverflow.com/a/29237343/101923
            cancel_event = Event()

            # Code to use multiprocess... would simplify exiting on stop, but progress doesn't currently work
            # with concurrent.futures.ProcessPoolExecutor(max_workers=len(read_session["streams"]), mp_context=mp.get_context('fork')) as executor:

            # See https://github.com/googleapis/python-bigquery/blob/main/google/cloud/bigquery/_pandas_helpers.py#L920
            if len(read_session["streams"]):
                with concurrent.futures.ThreadPoolExecutor(
                    max_workers=min(max_parallelization, len(read_session["streams"]))
                ) as executor:
                    futures = [
                        executor.submit(
                            process_stream,
                            stream,
                            folder_path,
                            mapped_variables,
                            coerce_schema,
                            progressbar,
                            batch_preprocessor,
                            cancel_event,
                        )
                        for stream in read_session["streams"]
                    ]

                    not_done = futures

                    try:
                        while not_done and not cancel_event.is_set():
                            # next line 'sleeps' this main thread, letting the thread pool run
                            freshly_done, not_done = concurrent.futures.wait(
                                not_done, timeout=0.2
                            )
                            for future in freshly_done:
                                # Call result() on any finished threads to raise any exceptions encountered.
                                future.result()
                    finally:
                        cancel_event.set()
                        # Shutdown all background threads, now that they should know to exit early.
                        executor.shutdown(wait=True, cancel_futures=True)

            if progressbar:
                progressbar.close()
            arrow_dataset = pyarrow_dataset.dataset(
                folder_path,
                format="feather",
                schema=(
                    pyarrow.schema(map(variable_to_field, mapped_variables))
                    if batch_preprocessor is None
                    else None
                ),
            )

        if output_type == "arrow_dataset":
            return arrow_dataset
        elif output_type == "polars_lazyframe":
            import polars

            return polars.scan_ipc(f"{folder_path}/*", memory_map=True)
        elif output_type == "dask_dataframe":
            import dask.dataframe as dd

            # TODO: simplify once dask supports reading from feather: https://github.com/dask/dask/issues/6865
            # Make sure we no longer remove the folder in the finally clause after making this change
            # Create the Parquet base directory
            parquet_base_dir = str(
                pathlib.Path()
                .joinpath(
                    get_tempdir(),
                    "tables",
                    f"{uuid.uuid4()}",
                )
                .absolute()
            )
            pyarrow_dataset.write_dataset(
                arrow_dataset, parquet_base_dir, format="parquet"
            )
            return dd.read_parquet(parquet_base_dir, dtype_backend="pyarrow")
        else:
            arrow_table = arrow_dataset.to_table()

            if output_type == "arrow_table":
                return arrow_table
            elif output_type == "tuple":
                variable_name_to_index = {}
                for index, variable in enumerate(mapped_variables):
                    variable_name_to_index[variable["name"]] = index

                pydict = arrow_table.to_pydict()
                keys = list(pydict.keys())

                return [
                    Row(
                        [pydict[variable["name"]][i] for variable in mapped_variables],
                        variable_name_to_index,
                    )
                    for i in range(len(pydict[keys[0]]))
                ]
    finally:
        if output_type != "arrow_dataset" and output_type != "polars_lazyframe":
            shutil.rmtree(folder_path, ignore_errors=True)


def variable_to_field(variable):
    import pyarrow

    if variable["type"] == "string" or variable["type"] == "geography":
        return pyarrow.field(variable["name"], pyarrow.string())
    elif variable["type"] == "integer":
        return pyarrow.field(variable["name"], pyarrow.int64())
    elif variable["type"] == "float":
        return pyarrow.field(variable["name"], pyarrow.float64())
    elif variable["type"] == "date":
        return pyarrow.field(variable["name"], pyarrow.date32())
    elif variable["type"] == "dateTime":
        return pyarrow.field(variable["name"], pyarrow.timestamp("us"))
    elif variable["type"] == "time":
        return pyarrow.field(variable["name"], pyarrow.time64("us"))
    elif variable["type"] == "boolean":
        return pyarrow.field(variable["name"], pyarrow.bool_())


def coerce_string_variable(pyarrow_array, variable):
    import pyarrow

    if variable["type"] == "string" or variable["type"] == "geography":
        return pyarrow_array
    elif variable["type"] == "integer":
        return pyarrow.compute.cast(pyarrow_array, pyarrow.int64())
    elif variable["type"] == "float":
        return pyarrow.compute.cast(pyarrow_array, pyarrow.float64())
    elif variable["type"] == "date":
        return pyarrow.compute.cast(
            pyarrow.compute.cast(pyarrow_array, pyarrow.timestamp("us")),
            pyarrow.date32(),
        )
    elif variable["type"] == "dateTime":
        return pyarrow.compute.cast(pyarrow_array, pyarrow.timestamp("us"))
    elif variable["type"] == "time":
        return pyarrow.compute.cast(
            pyarrow.compute.cast(
                pyarrow.compute.utf8_replace_slice(
                    pyarrow_array, start=0, stop=0, replacement="2020-01-01T"
                ),
                pyarrow.timestamp("us"),
            ),
            pyarrow.time64("us"),
        )
    elif variable["type"] == "boolean":
        return pyarrow.compute.cast(pyarrow_array, pyarrow.bool_())


def process_stream(
    stream,
    folder_path,
    mapped_variables,
    coerce_schema,
    progressbar,
    batch_preprocessor,
    cancel_event,
):
    import pyarrow

    with closing(
        make_request(
            method="get",
            path=f'/readStreams/{stream["id"]}',
            stream=True,
            parse_response=False,
        )
    ) as arrow_response:
        has_content = False
        # create the os_file path
        os_file = (
            pathlib.Path(folder_path).joinpath(f"{stream['id']}.feather").absolute()
        )
        with pyarrow.OSFile(
            str(os_file), mode="wb"
        ) as f, pyarrow.ipc.RecordBatchStreamReader(arrow_response.raw) as reader:
            if coerce_schema:
                variables_in_stream = list(
                    map(
                        lambda field_name: next(
                            x
                            for x in mapped_variables
                            if x["name"].lower() == field_name.lower()
                        ),
                        reader.schema.names,
                    )
                )

                output_schema = pyarrow.schema(
                    map(variable_to_field, variables_in_stream)
                )
            else:
                output_schema = reader.schema
            writer = None
            for batch in reader:
                # exit out of thread
                if cancel_event.is_set():
                    has_content = False
                    break

                if coerce_schema:
                    batch = pyarrow.RecordBatch.from_arrays(
                        list(
                            map(
                                coerce_string_variable,
                                batch.columns,
                                variables_in_stream,
                            )
                        ),
                        schema=output_schema,
                    )

                num_rows = batch.num_rows
                if batch_preprocessor:
                    batch = batch_preprocessor(batch)

                if batch is not None:
                    has_content = True
                    if writer is None:
                        writer = pyarrow.ipc.RecordBatchFileWriter(
                            f,
                            (
                                output_schema
                                if batch_preprocessor is None
                                else batch.schema
                            ),
                        )

                    writer.write_batch(batch)

                if progressbar is not None:
                    progressbar.update(num_rows)

            if writer is not None:
                writer.close()

        if has_content == False:
            os.remove(os_file)


def format_tuple_type(val, type):
    if val is None:
        return val
    elif type == "integer":
        return int(val)
    elif type == "float":
        return float(val)
    elif type == "date":
        return str(val)
    elif type == "dateTime":
        return str(val)
    elif type == "time":
        return str(val)
    elif type == "boolean":
        return bool(val)
    else:
        return str(val)



from tqdm.auto import tqdm
import os
from io import BytesIO
import pathlib
import re
import urllib
from .Base import Base
from ..common.api_request import make_request
from urllib.parse import quote as quote_uri


class File(Base):
    def __init__(
        self,
        id,
        *,
        table=None,
        properties={},
    ):
        self.id = id
        self.table = table
        self.uri = f"/rawFiles/{quote_uri(id, '')}"
        self.properties = {
            **{"kind": "rawFile", "id": id, "uri": self.uri},
            **properties,
        }

    def get(self):
        res = make_request(method="HEAD", path=self.uri, parse_response=False)
        parse_headers(self, res)
        return self

    def download(
        self,
        path=None,
        *,
        overwrite=False,
        progress=True,
        on_progress=None,
        cancel_event=None,
    ):
        is_dir = False
        if path is None:
            path = os.getcwd()
            is_dir = True
        elif path.endswith(os.sep):
            is_dir = True
        elif os.path.exists(path) and os.path.isdir(path):
            is_dir = True

        with make_request(
            method="GET",
            path=f"{self.uri}",
            query={"allowRedirect": "true"},
            stream=True,
            parse_response=False,
        ) as r:
            parse_headers(self, r)
            name = self.properties["name"]

            file_name = os.path.join(path, name) if is_dir else path

            if overwrite is False and os.path.exists(file_name):
                raise Exception(
                    f"File already exists at '{file_name}'. Set parameter overwrite=True to overwrite existing files."
                )

            # Make sure output directory exists
            pathlib.Path(file_name).parent.mkdir(exist_ok=True, parents=True)

            with open(file_name, "wb") as f:
                if progress:
                    pbar = tqdm(
                        total=self.properties["size"],
                        leave=False,
                        unit="iB",
                        unit_scale=True,
                    )
                for chunk in r.iter_content(chunk_size=1024 * 1024):
                    if cancel_event and cancel_event.is_set():
                        os.remove(file_name)
                        return None
                    if progress:
                        pbar.update(len(chunk))
                    f.write(chunk)
                    if on_progress:
                        on_progress(len(chunk))

                if progress:
                    pbar.close()

        return file_name

    def read(self, *, as_text=False):
        r = make_request(method="GET", path=f"{self.uri}", parse_response=False)
        parse_headers(self, r)
        if as_text:
            return r.text
        else:
            return r.content

    def stream(self):
        r = make_request(
            method="GET", path=f"{self.uri}", parse_response=False, stream=True
        )
        parse_headers(self, r)
        return BytesIO(r.content)


def parse_headers(file, res):
    file.properties["name"] = get_filename(res.headers["content-disposition"])
    file.properties["contentType"] = res.headers["content-type"]

    digest = None

    if "Digest" in res.headers:
        digest = res.headers["Digest"]
    elif "x-goog-hash" in res.headers:
        digest = res.headers["x-goog-hash"]

    if digest:
        file.properties["md5"] = digest.replace("md5=", "")

    file.properties["size"] = int(
        res.headers["content-length"] or res.headers["x-goog-stored-content-length"]
    )


def get_filename(s):
    fname = re.findall("filename\*=([^;]+)", s, flags=re.IGNORECASE)
    if not fname:
        fname = re.findall("filename=([^;]+)", s, flags=re.IGNORECASE)
    if "utf-8''" in fname[0].lower():
        fname = re.sub("utf-8''", "", fname[0], flags=re.IGNORECASE)
        fname = urllib.unquote(fname).decode("utf8")
    else:
        fname = fname[0]
    # clean space and double quotes
    return fname.strip().strip('"')



import time
from .Base import Base
from ..common.api_request import make_request


class Variable(Base):
    def __init__(
        self,
        name,
        *,
        table=None,
        upload=None,
        properties=None,
    ):
        self.name = name
        self.table = table
        self.upload = upload
        self.uri = (
            properties["uri"]
            if "uri" in (properties or {})
            else (f"{table.uri if table else upload.uri}/variables/{self.name}")
        )
        self.properties = properties

    def get(self, wait_for_statistics=False):
        self.properties = make_request(
            method="GET",
            path=self.uri,
        )
        self.uri = self.properties["uri"]
        while (
            wait_for_statistics and self.properties["statistics"]["status"] == "running"
        ):
            time.sleep(2)
            self.properties = make_request(
                method="GET",
                path=self.uri,
            )

        return self

    def update(self, *, label=None, description=None, value_labels=None):
        payload = {}
        if label is not None:
            payload["label"] = label
        if description is not None:
            payload["description"] = description
        if value_labels is not None:
            payload["valueLabels"] = value_labels

        response = make_request(
            method="PATCH",
            path=f"{self.uri}",
            payload=payload,
        )
        self.properties = response
        return

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except Exception as err:
            if err.args[0]["status"] != 404:
                raise err
            return False



import copy
from typing import Any, Dict, Iterable, Tuple


class Row(object):
    """Represents a row in a table or query result.
    Values can be accessed by position (index), by key like a dict,
    or as properties.
    Args:
        values (Sequence[object]): The row values
        variable_name_to_index (Dict[str, int]):
            A mapping from variable names to indexes
    """

    # Avoid conflict with schema fields.
    __slots__ = ("__redivis_hidden_values", "__redivis_hidden_variable_name_to_index")

    def __init__(self, values, variable_name_to_index) -> None:
        self.__redivis_hidden_values = values
        self.__redivis_hidden_variable_name_to_index = variable_name_to_index

    def values(self):
        """Return the values included in this row.
        Returns:
            Sequence[object]: A sequence of length ``len(row)``.
        """
        return copy.deepcopy(self.__redivis_hidden_values)

    def keys(self) -> Iterable[str]:
        """Return the keys for using a row as a dict.
        Returns:
            Iterable[str]: The keys corresponding to the columns of a row
        """
        return self.__redivis_hidden_variable_name_to_index.keys()

    def items(self) -> Iterable[Tuple[str, Any]]:
        """Return items as ``(key, value)`` pairs.
        Returns:
            Iterable[Tuple[str, object]]:
                The ``(key, value)`` pairs representing this row.
        """
        for key, index in self.__redivis_hidden_variable_name_to_index.items():
            yield key, self.__redivis_hidden_values[index]

    def get(self, key: str) -> Any:
        """Return a value for key, with a default value if it does not exist.
        Args:
            key (str): The key of the column to access
        Returns:
            object:
                The value associated with the provided key
        """
        index = self.__redivis_hidden_variable_name_to_index.get(key)
        if index is None:
            raise KeyError(f"Row does not contain field {key}")
        return self.__redivis_hidden_values[index]

    def __getattr__(self, name):
        value = self.__redivis_hidden_variable_name_to_index.get(name)
        if value is None:
            raise AttributeError(f"Row does not contain field {name}")
        return self.__redivis_hidden_values[value]

    def __len__(self):
        return len(self.__redivis_hidden_values)

    def __getitem__(self, key):
        if isinstance(key, str):
            value = self.__redivis_hidden_variable_name_to_index.get(key)
            if value is None:
                raise KeyError(f"Row does not contain field {key}")
            key = value
        return self.__redivis_hidden_values[key]

    def __eq__(self, other):
        if not isinstance(other, Row):
            return NotImplemented
        return (
            self.__redivis_hidden_values == other.__redivis_hidden_values
            and self.__redivis_hidden_variable_name_to_index
            == other.__redivis_hidden_variable_name_to_index
        )

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        column_index_to_variable_name = {}
        for variable_name in self.__redivis_hidden_variable_name_to_index.keys():
            column_index_to_variable_name[
                self.__redivis_hidden_variable_name_to_index[variable_name]
            ] = variable_name
        cells = []
        for index, value in enumerate(self.__redivis_hidden_values):
            cells.append(f"{column_index_to_variable_name[index]}: {value}")

        f2i = "{" + ", ".join(cells) + "}"
        return f"Row({f2i})"



from .classes.User import User as user
from .classes.Organization import Organization as organization
from .classes.Query import Query as query
from .classes.Table import Table as table
from .classes.File import File as file
import warnings

from ._version import __version__


def _warning(message, category, filename, lineno, file=None, line=None):
    print(f"Warning: {message}")


warnings.showwarning = _warning


def authenticate(scope=None, force_reauthentication=False):
    from .common.auth import clear_cached_credentials, get_auth_token

    if force_reauthentication:
        clear_cached_credentials()

    if isinstance(scope, str):
        scope = [scope]

    get_auth_token(scope=scope)


def current_notebook():
    import os

    if os.getenv("REDIVIS_NOTEBOOK_JOB_ID") is not None:
        from .classes.Notebook import Notebook

        return Notebook(os.getenv("REDIVIS_NOTEBOOK_JOB_ID"))


import pkg_resources  # part of setuptools



import util


def test_list_variables():
    util.populate_test_data()
    table = util.get_table()
    variables = table.list_variables()
    print(variables)


def test_list_rows():
    util.populate_test_data()
    table = util.get_table()
    print(table.list_rows(max_results=10))


def test_to_dataframe():
    util.populate_test_data()
    table = util.get_table()
    print(table.to_dataframe(max_results=10))


def test_selected_variables():
    util.populate_test_data()
    table = util.get_table()
    print(table.to_dataframe(max_results=10, variables=["ID"]))


def test_geo_dataframe():
    util.delete_test_dataset()
    util.populate_test_data("us_states.geojsonl")
    table = util.get_table()
    df = table.to_dataframe(max_results=100)
    print(df.dtypes)
    print(df)


import re
import math
import time
import requests
import os
import logging
from tqdm.utils import CallbackIOWrapper
from urllib.parse import quote as quote_uri

from .auth import get_auth_token

verify_ssl = (
    False
    if os.getenv("REDIVIS_API_ENDPOINT", "https://redivis.com/api/v1").find(
        "https://localhost", 0
    )
    == 0
    else True
)


def perform_resumable_upload(
    data, size=None, temp_upload_url=None, proxy_url=None, progressbar=None
):
    retry_count = 0
    start_byte = 0
    is_file = True if hasattr(data, "read") else False
    file_size = size

    if file_size is None:
        file_size = os.stat(data.name).st_size if is_file else len(data)

    chunk_size = file_size
    # chunk_size = min(file_size, 2**30)
    headers = {"Authorization": f"Bearer {get_auth_token()}"}

    if proxy_url:
        temp_upload_url = f"{proxy_url}?url={quote_uri(temp_upload_url)}"

    resumable_url = initiate_resumable_upload(file_size, temp_upload_url, headers)

    while (
        start_byte < file_size or start_byte == 0
    ):  # handle empty upload for start_byte == 0
        end_byte = min(start_byte + chunk_size - 1, file_size - 1)
        if progressbar:
            progressbar.update(start_byte - progressbar.n)
        if is_file:
            data.seek(start_byte)
            if progressbar:
                chunk = CallbackIOWrapper(progressbar.update, data, "read")
            else:
                chunk = data
        else:
            if start_byte != 0:
                chunk = data[start_byte : end_byte + 1]
            else:
                chunk = data

        try:
            res = requests.put(
                url=resumable_url,
                verify=verify_ssl,
                headers={
                    **headers,
                    **{
                        "Content-Length": f"{end_byte - start_byte + 1}",
                        "Content-Range": f"bytes {start_byte}-{end_byte}/{file_size}",
                    },
                },
                data=chunk,
            )
            res.raise_for_status()

            start_byte += chunk_size
            retry_count = 0  # reset retry_count after a successfully uploaded chunk
        except Exception as e:
            if retry_count > 10:
                print("A network error occurred. Upload failed after too many retries.")
                raise e

            retry_count += 1
            time.sleep(retry_count)
            print("A network error occurred. Retrying last chunk of resumable upload.")
            start_byte = retry_partial_upload(
                file_size=file_size, resumable_url=resumable_url, headers=headers
            )


def initiate_resumable_upload(size, temp_upload_url, headers, retry_count=0):
    did_request_complete = False
    try:
        res = requests.post(
            temp_upload_url,
            verify=verify_ssl,
            headers={
                **headers,
                **{
                    "x-upload-content-length": str(size),
                    "x-goog-resumable": "start",
                },
            },
        )
        did_request_complete = True
        res.raise_for_status()
        return res.headers["location"]

    except Exception as e:
        if did_request_complete:
            raise e
        else:
            if retry_count > 10:
                print("A network error occurred. Upload failed after too many retries.")
                raise e
            time.sleep(retry_count + 1)
            initiate_resumable_upload(
                size, temp_upload_url, headers, retry_count=retry_count + 1
            )


def retry_partial_upload(*, retry_count=0, file_size, resumable_url, headers):
    logging.debug("Attempting to resume upload")

    try:
        res = requests.put(
            url=resumable_url,
            verify=verify_ssl,
            headers={
                **headers,
                **{"Content-Length": "0", "Content-Range": f"bytes */{file_size}"},
            },
        )

        if res.status_code == 404:
            return 0

        res.raise_for_status()

        if res.status_code == 200 or res.status_code == 201:
            return file_size
        elif res.status_code == 308:
            range_header = res.headers["Range"] if "Range" in res.headers else None
            if range_header:
                match = re.match(r"bytes=0-(\d+)", range_header)
                if match.group(0) and not math.isnan(int(match.group(1))):
                    return int(match.group(1)) + 1
                else:
                    raise Exception("An unknown error occurred. Please try again.")
            # If GCS hasn't received any bytes, the header will be missing
            else:
                return 0
        else:
            raise Exception("An unknown error occurred. Please try again.")
    except Exception as e:
        if retry_count > 10:
            raise e

        time.sleep(retry_count + 1)
        return retry_partial_upload(
            retry_count=retry_count + 1,
            file_size=file_size,
            resumable_url=resumable_url,
            headers=headers,
        )


def perform_standard_upload(
    data, temp_upload_url=None, proxy_url=None, retry_count=0, progressbar=None
):
    original_url = temp_upload_url
    try:
        if progressbar:
            data = CallbackIOWrapper(progressbar.update, data, "read")

        headers = {"Authorization": f"Bearer {get_auth_token()}"}

        if proxy_url:
            temp_upload_url = f"{proxy_url}?url={quote_uri(temp_upload_url)}"

        res = requests.put(
            url=temp_upload_url, data=data, headers=headers, verify=verify_ssl
        )
        res.raise_for_status()
    except Exception as e:
        if retry_count > 10:
            print("A network error occurred. Upload failed after too many retries.")
            raise e
        time.sleep(retry_count + 1)
        return perform_standard_upload(
            data=data,
            temp_upload_url=original_url,
            proxy_url=proxy_url,
            retry_count=retry_count + 1,
            progressbar=progressbar,
        )



from .Base import Base
import os
import time
import concurrent.futures
import pathlib
from ..common.api_request import make_request
from tqdm.auto import tqdm
import re
import urllib
from threading import Event
from contextlib import closing


class Export(Base):
    def __init__(
        self,
        id,
        *,
        table=None,
        properties={},
    ):
        self.table = table
        self.properties = properties
        self.uri = self.properties["uri"]

    def get(self):
        self.properties = make_request(method="GET", path=self.uri)
        return self

    def download_files(self, *, path, overwrite, progress=True):
        # TODO: if overwriting file, first check if file is the same size, and if so check md5 hash, and skip if identical (need md5 in header)
        self.wait_for_finish()
        file_count = self.properties["fileCount"]
        is_dir = False
        if path is None or (os.path.exists(path) and os.path.isdir(path)):
            is_dir = True
            if path is None:
                path = os.getcwd()
            if file_count > 1:
                if not hasattr(self.table.properties, "name"):
                    self.table.get()
                escaped_table_name = re.sub(
                    r"\W+", "_", self.table.properties["name"]
                ).lower()
                path = os.path.join(path, escaped_table_name)
        elif path.endswith(os.sep) or (not os.path.exists(path) and "." not in path):
            is_dir = True
        elif file_count > 1:
            raise Exception(
                f"Path '{path}' is a file, but the export consists of multiple files. Please specify the path to a directory"
            )

        if (
            overwrite is False
            and os.path.exists(path)
            and (not is_dir or file_count > 1)
        ):
            raise Exception(
                f"File already exists at '{path}'. Set parameter overwrite=True to overwrite existing files."
            )

        # Make sure output directory exists
        if is_dir:
            pathlib.Path(path).mkdir(exist_ok=True, parents=True)
        else:
            pathlib.Path(path).parent.mkdir(exist_ok=True, parents=True)

        pbar = None
        if progress:
            pbar = tqdm(
                total=self.properties["size"], leave=False, unit="iB", unit_scale=True
            )

        cancel_event = Event()
        # Code to use multiprocess... would simplify exiting on stop, but progress doesn't currently work
        # with concurrent.futures.ProcessPoolExecutor(max_workers=len(read_session["streams"]), mp_context=mp.get_context('fork')) as executor:

        output_file_paths = []
        # TODO: this should use async, not threads
        # See https://github.com/googleapis/python-bigquery/blob/main/google/cloud/bigquery/_pandas_helpers.py#L920
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=min(os.cpu_count() * 5, file_count)
        ) as executor:
            futures = [
                executor.submit(
                    download_file,
                    uri=f"{self.uri}/download",
                    download_path=path,
                    file_number=file_number,
                    is_dir=is_dir,
                    overwrite=overwrite,
                    pbar=pbar,
                    cancel_event=cancel_event,
                )
                for file_number in range(file_count)
            ]

            not_done = futures

            try:
                while not_done and not cancel_event.is_set():
                    # next line 'sleeps' this main thread, letting the thread pool run
                    freshly_done, not_done = concurrent.futures.wait(
                        not_done, timeout=0.2
                    )
                    for future in freshly_done:
                        # Call result() on any finished threads to raise any exceptions encountered.
                        output_file_paths.append(future.result())
            finally:
                cancel_event.set()
                # Shutdown all background threads, now that they should know to exit early.
                executor.shutdown(wait=True, cancel_futures=True)

        if pbar:
            pbar.close()

        return output_file_paths

    def wait_for_finish(self, *, progress=True):
        iter_count = 0
        if progress:
            pbar = tqdm(total=100, leave=False, unit="%", unit_scale=True)
            pbar.set_description(f"Preparing download...")

        while True:
            if self.properties["status"] == "completed":
                if progress:
                    pbar.close()
                break
            elif self.properties["status"] == "failed":
                if progress:
                    pbar.close()
                raise Exception(
                    f"Export job failed with message: {self.properties['errorMessage']}"
                )
            elif self.properties["status"] == "cancelled":
                if progress:
                    pbar.close()
                raise Exception(f"Export job was cancelled")
            else:
                iter_count += 1
                if progress:
                    pbar.update(self.properties["percentCompleted"] - pbar.n)
                time.sleep(min(iter_count * 0.5, 2))
                self.get()


def get_filename(s):
    fname = re.findall("filename\*=([^;]+)", s, flags=re.IGNORECASE)
    if not fname:
        fname = re.findall("filename=([^;]+)", s, flags=re.IGNORECASE)
    if "utf-8''" in fname[0].lower():
        fname = re.sub("utf-8''", "", fname[0], flags=re.IGNORECASE)
        fname = urllib.unquote(fname).decode("utf8")
    else:
        fname = fname[0]
    # clean space and double quotes
    return fname.strip().strip('"')


def download_file(
    *,
    uri,
    file_number,
    download_path,
    is_dir=False,
    overwrite=False,
    pbar=None,
    cancel_event,
):
    with closing(
        make_request(
            method="GET",
            path=uri,
            query={"filePart": file_number},
            stream=True,
            parse_response=False,
        )
    ) as r:
        if is_dir:
            file_name = get_filename(r.headers.get("Content-Disposition"))
            download_path = os.path.join(download_path, file_name)

        if overwrite is False and os.path.exists(download_path):
            raise Exception(
                f"File already exists at '{download_path}'. Set parameter overwrite=True to overwrite existing files."
            )

        with open(download_path, "wb") as f:
            for chunk in r.iter_content(chunk_size=1024 * 1024):
                if cancel_event.is_set():
                    break
                if pbar:
                    pbar.update(len(chunk))
                f.write(chunk)

    return download_path



from urllib.parse import quote as quote_uri
from .Dataset import Dataset
from .Base import Base
from .Workflow import Workflow
from ..common.api_request import make_paginated_request
import warnings


class User(Base):
    def __init__(self, name, properties={}):
        self.name = name
        self.uri = f"/users/{quote_uri(self.name, '')}"
        self.properties = {
            **{"kind": "user", "userName": name, "uri": self.uri},
            **properties,
        }

    def dataset(self, name, *, version=None):
        return Dataset(name, user=self, version=version)

    def project(self, name):
        warnings.warn(
            "Projects have been renamed to Workflows, please update your code to: user.workflow()",
            FutureWarning,
            stacklevel=2,
        )
        return Workflow(name, user=self)

    def workflow(self, name):
        return Workflow(name, user=self)

    def list_datasets(self, max_results=None):
        datasets = make_paginated_request(
            path=f"{self.uri}/datasets", page_size=100, max_results=max_results
        )
        return [
            Dataset(dataset["name"], user=self, properties=dataset)
            for dataset in datasets
        ]

    def list_projects(self, max_results=None):
        warnings.warn(
            "Projects have been renamed to Workflows, please update your code to: user.list_workflows()",
            FutureWarning,
            stacklevel=2,
        )
        return self.list_workflows(max_results)

    def list_workflows(self, max_results=None):
        workflows = make_paginated_request(
            path=f"{self.uri}/workflows", page_size=100, max_results=max_results
        )
        return [
            Workflow(workflow["name"], user=self, properties=workflow)
            for workflow in workflows
        ]



from .Table import Table
from .Base import Base
from ..common.api_request import make_request
from tqdm.auto import tqdm

import pathlib
import os

from ..common.retryable_upload import perform_resumable_upload, perform_standard_upload
from ..common.util import convert_data_to_parquet


class Notebook(Base):
    def __init__(
        self,
        current_notebook_job_id,
    ):
        self.current_notebook_job_id = current_notebook_job_id

    def create_output_table(
        self, data=None, *, name=None, append=False, geography_variables=None
    ):
        should_remove_tempfile = True
        temp_file_path = None
        try:
            if isinstance(data, str) or isinstance(data, pathlib.PurePath):
                data = str(data)
                if data.endswith(".parquet"):
                    should_remove_tempfile = False
                    temp_file_path = data
                else:
                    raise Exception(
                        "Only paths to parquet files (ending in .parquet) are supported when a string argument is provided"
                    )
            else:
                temp_file_path = convert_data_to_parquet(data)

            size = os.stat(temp_file_path).st_size
            pbar_bytes = tqdm(total=size, unit="B", leave=False, unit_scale=True)

            res = make_request(
                method="POST",
                path=f"/notebookJobs/{self.current_notebook_job_id}/tempUploads",
                payload={"tempUploads": [{"size": size, "resumable": size > 5e7}]},
            )
            temp_upload = res["results"][0]

            with open(temp_file_path, "rb") as f:
                if temp_upload["resumable"]:
                    perform_resumable_upload(
                        data=f,
                        progressbar=pbar_bytes,
                        proxy_url=f"{os.getenv('REDIVIS_API_ENDPOINT')}/notebookJobs/{self.current_notebook_job_id}/tempUploadProxy",
                        temp_upload_url=temp_upload["url"],
                    )
                else:
                    perform_standard_upload(
                        data=f,
                        temp_upload_url=temp_upload["url"],
                        proxy_url=f"{os.getenv('REDIVIS_API_ENDPOINT')}/notebookJobs/{self.current_notebook_job_id}/tempUploadProxy",
                        progressbar=pbar_bytes,
                    )

            pbar_bytes.close()
            if should_remove_tempfile:
                os.remove(temp_file_path)

            res = make_request(
                method="PUT",
                path=f"/notebookJobs/{self.current_notebook_job_id}/outputTable",
                payload={
                    "name": name,
                    "append": append,
                    "geographyVariables": geography_variables,
                    "tempUploadId": temp_upload["id"],
                },
            )

            return Table(name=res["name"], properties=res)
        except Exception as e:
            if (
                temp_file_path
                and os.path.exists(temp_file_path)
                and should_remove_tempfile
            ):
                os.remove(temp_file_path)
            raise e



import os
import tempfile
import atexit
import shutil
import pathlib
import uuid


def get_geography_variable(variables, geography_variable_name):
    if geography_variable_name == "":
        for variable in variables:
            if variable["type"] == "geography":
                return variable

        return None
    else:
        for variable in variables:
            if variable["name"] == geography_variable_name:
                return variable

        raise Exception(
            f"The specified geography variable '{geography_variable_name()}' could not be found"
        )


def get_warning(kind):
    if kind == "dataframe_deprecation":
        return 'The to_dataframe() method is deprecated, and has been superceded by to_pandas_dataframe().\nBy default, this new method uses the "pyarrow" dtype_backend, which is more performant and will generally work with existing code.\nTo replicate historic behavior, use to_pandas_dataframe(dtype_backend="numpy").'
    elif kind == "geodataframe_deprecation":
        return "Please use the to_geopandas_dataframe() method to ensure future compatability."
    else:
        return "WARNING"


created_temp_dir = None


def rm_tempdir():
    global created_temp_dir
    shutil.rmtree(created_temp_dir, ignore_errors=True)


def get_tempdir():
    if os.getenv("REDIVIS_TMPDIR"):
        user_suffix = os.environ.get("USER", os.environ.get("USERNAME")) or os.getuid()
        return str(
            pathlib.Path("/").joinpath(
                os.getenv("REDIVIS_TMPDIR"),
                f"redivis_{user_suffix}",
            )
        )

    global created_temp_dir
    if created_temp_dir is None:
        created_temp_dir = tempfile.mkdtemp()
        atexit.register(rm_tempdir)

    return created_temp_dir


def convert_data_to_parquet(data):
    temp_file_path = f"{get_tempdir()}/parquet/{uuid.uuid4()}"
    pathlib.Path(temp_file_path).parent.mkdir(exist_ok=True, parents=True)

    import geopandas
    import pandas as pd
    import pyarrow as pa
    import pyarrow.dataset as pa_dataset
    import pyarrow.parquet as pa_parquet
    from dask.dataframe import DataFrame as dask_df

    if isinstance(data, geopandas.GeoDataFrame):
        data.to_wkt().to_parquet(path=temp_file_path, index=False)
    elif isinstance(data, pd.DataFrame):
        data.to_parquet(path=temp_file_path, index=False)
    elif isinstance(data, pa_dataset.Dataset):
        pa_dataset.write_dataset(
            data,
            temp_file_path,
            format="parquet",
            basename_template="part-{i}.parquet",
            max_partitions=1,
        )
        temp_file_path = f"{temp_file_path}/part-0.parquet"
    elif isinstance(data, pa.Table):
        pa_parquet.write_table(data, temp_file_path)
    elif isinstance(data, dask_df):
        data.to_parquet(temp_file_path, write_index=False)
        temp_file_path = f"{temp_file_path}/part.0.parquet"
    else:
        # importing polars is causing an IllegalInstruction error on ARM + Docker. Import inline to avoid crashes elsewhwere
        # TODO: revert once fixed upstream
        import polars

        if isinstance(data, polars.LazyFrame):
            data.sink_parquet(temp_file_path)
        elif isinstance(data, polars.DataFrame):
            data.write_parquet(temp_file_path)
        else:
            raise Exception(
                "Unknown datatype provided. Must be an instance of pandas.DataFrame, pyarrow.Dataset, pyarrow.Table, dask.DataFrame, polars.LazyFrame, or polars.DataFrame"
            )

    return temp_file_path


def arrow_table_to_pandas(
    arrow_table, dtype_backend, date_as_object, max_parallelization
):
    import pandas as pd
    import pyarrow as pa

    pa.set_cpu_count(max_parallelization)
    pa.set_io_thread_count(max_parallelization)

    if dtype_backend not in ["numpy", "numpy_nullable", "pyarrow"]:
        raise Exception(
            f"Unknown dtype_backend. Must be one of 'pyarrow'|'numpy_nullable'|'numpy'. Default is 'pyarrow'"
        )

    if dtype_backend == "numpy_nullable":
        df = arrow_table.to_pandas(
            self_destruct=True,
            date_as_object=date_as_object,
            types_mapper={
                pa.int64(): pd.Int64Dtype(),
                pa.bool_(): pd.BooleanDtype(),
                pa.float64(): pd.Float64Dtype(),
                pa.string(): pd.StringDtype(),
            }.get,
        )
    elif dtype_backend == "pyarrow":
        df = arrow_table.to_pandas(self_destruct=True, types_mapper=pd.ArrowDtype)
    else:
        df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object)

    return df



from .Query import Query
from .Table import Table
from .Base import Base
from urllib.parse import quote as quote_uri
from ..common.api_request import make_request, make_paginated_request


class Workflow(Base):
    def __init__(self, name, *, user, properties=None):
        self.user = user
        self.name = name

        self.qualified_reference = (
            properties["qualifiedReference"]
            if "qualifiedReference" in (properties or {})
            else (f"{self.user.name}.{self.name}")
        )
        self.scoped_reference = (
            properties["scopedReference"]
            if "scopedReference" in (properties or {})
            else f"{self.name}"
        )
        self.uri = f"/workflows/{quote_uri(self.qualified_reference, '')}"
        self.properties = properties

    def list_tables(self, *, max_results=None):
        tables = make_paginated_request(
            path=f"{self.uri}/tables",
            page_size=100,
            max_results=max_results,
        )
        return [
            Table(table["name"], workflow=self, properties=table) for table in tables
        ]

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except Exception as err:
            if err.args[0]["status"] != 404:
                raise err
            return False

    def get(self):
        properties = make_request(method="GET", path=self.uri)
        update_properties(self, properties)
        return self

    def query(self, query):
        return Query(query, default_workflow=self.qualified_reference)

    def table(self, name):
        return Table(name, workflow=self)


def update_properties(instance, properties):
    instance.properties = properties
    instance.qualified_reference = properties["qualifiedReference"]
    instance.scoped_reference = properties["scopedReference"]
    instance.name = properties["name"]
    instance.uri = properties["uri"]



import time
from .Base import Base
from .User import User
from ..common.api_request import make_request


class Member(Base):
    def __init__(self, name, *, organization, properties={}):
        self.name = name
        self.organization = organization
        self.user = User(
            name, properties=properties["user"] if "user" in properties else {}
        )
        self.uri = (
            properties["uri"]
            if "uri" in (properties or {})
            else (f"{organization.uri}/members/{self.name}")
        )
        self.properties = properties

    def get(self):
        self.properties = make_request(
            method="GET",
            path=self.uri,
        )
        self.uri = self.properties["uri"]
        self.user.properties = self.properties["user"]
        return self

    def add_labels(self, labels):
        self.get()
        self.update(
            labels=list(
                set(self.properties.get("labels", []))
                | set(label.lower() for label in labels)
            )
        )
        return self

    def remove_labels(self, labels):
        self.get()
        self.update(
            labels=list(
                set(self.properties.get("labels", []))
                - set(label.lower() for label in labels)
            )
        )
        return self

    def update(self, *, labels=None, status=None):
        payload = {}
        if labels is not None:
            payload["labels"] = labels
        if status is not None:
            payload["status"] = status

        response = make_request(
            method="PATCH",
            path=f"{self.uri}",
            payload=payload,
        )
        self.properties = response
        self.user.properties = response["user"]
        return

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except Exception as err:
            if err.args[0]["status"] != 404:
                raise err
            return False
